import numpy as np
from PIL import Image

# Define image size and camera parameters
img_width = 640
img_height = 480
fov = np.pi/2
aspect_ratio = img_width/img_height
eye_pos = np.array([0, 0, 0])
z_dir = np.array([0, 0, 1])

# Define sphere parameters
sphere_pos = np.array([0, 0, 3])
sphere_radius = 1
sphere_color = np.array([1, 0, 0])

# Define lighting parameters
light_pos = np.array([0, 10, 10])
light_color = np.array([1, 1, 1])

# Initialize image buffer
img_buf = np.zeros((img_height, img_width, 3))

# Ray-sphere intersection test
def intersect_sphere(ray_origin, ray_direction, center, radius):
    b = 2 * np.dot(ray_direction, ray_origin - center)
    c = np.linalg.norm(ray_origin - center) ** 2 - radius ** 2
    delta = b ** 2 - 4 * c
    if delta > 0:
        t1 = (-b + np.sqrt(delta)) / 2
        t2 = (-b - np.sqrt(delta)) / 2
        if t1 > 0 and t2 > 0:
            return min(t1, t2)
    return None

# Trace a ray and return the color
def trace_ray(ray_orig, ray_dir):
    # Find closest intersection with sphere
    t = intersect_sphere(ray_orig, ray_dir, sphere_pos, sphere_radius)
    if t is None:
        # print('none')
        return np.zeros(3)
    else:
        print(t)
    # Calculate intersection point and normal
    p = ray_orig + t*ray_dir
    n = (p - sphere_pos) / sphere_radius
    # Calculate diffuse lighting
    l = light_pos - p
    l /= np.linalg.norm(l)
    diffuse = max(np.dot(n, l), 0)
    # Calculate color
    color = sphere_color*diffuse*light_color
    return color

# Main rendering loop
for j in range(img_height):
    for i in range(img_width):
        # Calculate ray direction
        x = (2*(i + 0.5)/img_width - 1)*aspect_ratio*np.tan(fov/2)
        y = (1 - 2*(j + 0.5)/img_height)*np.tan(fov/2)
        ray_dir = np.array([x, y, -1])
        ray_dir /= np.linalg.norm(ray_dir)
        # Calculate camera direction
        right_dir = np.cross(ray_dir, z_dir)
        up_dir = np.cross(right_dir, ray_dir)
        cam_dir = ray_dir + 0.5*right_dir + 0.5*up_dir
        cam_dir /= np.linalg.norm(cam_dir)

        # Trace ray and set pixel color
        color = trace_ray(eye_pos, cam_dir)
        if not np.array_equal(color, np.array([0,0,0])):
            print(color)
        img_buf[j, i] = color

# Save image to file
img = Image.fromarray((255*img_buf).astype(np.uint8))
img.save('output.png')